import concurrent.futures
from timeit import default_timer
import numpy as np
import time
import math
from datetime import datetime

from karabo.bound import (
    Timestamp,
    Epochstamp,
    Trainstamp,
    ImageData,
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OVERWRITE_ELEMENT,
    NODE_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    IMAGEDATA_ELEMENT,
    UINT32_ELEMENT,
    ChannelMetaData,
    OUTPUT_CHANNEL,
    Hash,
    MetricPrefix,
    State,
    Schema,
    SLOT_ELEMENT,
    Unit,
)
from TrainMatcher import TrainMatcher

from calngUtils import shmem_utils, trackers

from onlinemid import agipd_module_name_from_mid_agipd_source, hash_from_dict

from onlinexpcsutils import onlinexpcsutils
import runningstats
    
def preview_schema():
    res = Schema()
    (

        NODE_ELEMENT(res)
        .key("data")
        .displayedName('Data')
        .commit(),
    
        IMAGEDATA_ELEMENT(res)
        .key("data.masking_plot")
        .displayedName('plots that show masking')
        .description('Masking Results')
        .commit(),
    
        IMAGEDATA_ELEMENT(res)
        .key("data.accumulating_sum")
        .displayedName('accumulating sum')
        .description('accumulating sum of measure pixel values, photonized between 0 and 50 or <high value> otherwise')
        .commit(),

        IMAGEDATA_ELEMENT(res)
        .key("data.accumulating_lit_bunch")
        .displayedName('accumulating lit bunch')
        .description('accumulating lit bunch pixel counter')
        .commit(),

        IMAGEDATA_ELEMENT(res)
        .key("data.current_det_image")
        .displayedName('accumulating lit bunch')
        .description('accumulating lit bunch pixel counter')
        .commit(),

        )
    return res



@KARABO_CLASSINFO("OnlineMIDProcessor", "0.0.1")
class OnlineMIDProcessor(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            
            BOOL_ELEMENT(expected)
            .key("enableKaraboOutput")
            .displayedName("Enable Karabo channel")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useInfiniband")
            .description(
                "If enabled, device will during initialization try to bind its data "
                "output channel (output) to its node's infiniband interface. Default "
                "interface is used if no 'ib0' interface is found."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),

            NODE_ELEMENT(expected)
            .key("debug")
            .displayname("Debug")
            .commit(),

            SLOT_ELEMENT(expected)
            .key("triggerSampleOutput")
            .displayedName("send test data over pipeline")
            .description("send test data")
            .commit(),


            OUTPUT_CHANNEL(expected)
            .key("plots")
            .dataSchema(preview_schema())
            .commit(),



        )
        
    def __init__(self, config):
        if config.get("useInfiniband", default=True):
            from PipeToZeroMQ.utils import find_infiniband_ip

            config["output.hostname"] = find_infiniband_ip()
        self._processing_time_tracker = trackers.ExponentialMovingAverage(alpha=0.3)
        super().__init__(config)
        self.info.merge(Hash("processingTime", 0))
        
        # expose the get xpcs settings function
        self.KARABO_SLOT(self.get_OnlineMIDSettings)
        self.KARABO_SLOT(self.triggerSampleOutput)
        self.registerSlot(self.triggerSampleOutput)

        ### XPCS specific variables
        self.online_processing_memory = None
        self.xpcs_processing_plot = None
        self.xpcs_settings = None

        # running stats specific variables
        self.runningstats_processing_memory = None
       
        self.timeSinceLastDataPrintUpdate = 0

    def initialization(self):
        super().initialization()
        self._shmem_handler = shmem_utils.ShmemCircularBufferReceiver()

        # initalize member variable for the q bin array 
        # (Will have shape of 16x512x128)
        self.assembled_output = self.signalSlotable.getOutputChannel("plots")
        
        if not self.get("enableKaraboOutput"):
            # it is set already by super by default, so only need to turn off
            self.output = None

        self.start()  # Auto-start this type of matcher.

    def preReconfigure(self, conf):
        super().preReconfigure(conf)

        if conf.has("enableKaraboOutput"):
            if conf["enableKaraboOutput"]:
                self.output = self._ss.getOutputChannel("output")
            else:
                self.output = None
        
    def get_OnlineMIDSettings(self, onlineAnalysisSettings_as_Hash):
        

        '''
        self.runningstats_processing_memory = None

                
        xpcs_settings = dict()
        

        xpcs_settings['qbin'] = xpcs_settings_as_hash['qbin']
        xpcs_settings['adapting_mask'] = xpcs_settings_as_hash['adapting_mask']
        xpcs_settings['adapting_mask_idx'] = xpcs_settings_as_hash['adapting_mask_idx']
        xpcs_settings['adapting_mask_idy'] = xpcs_settings_as_hash['adapting_mask_idy']
        xpcs_settings['adapting_mask_block_indeces'] = xpcs_settings_as_hash['adapting_mask_block_indeces']
        xpcs_settings['single_image_masking'] = xpcs_settings_as_hash['single_image_masking']
        xpcs_settings['single_image_masking_percentile'] = xpcs_settings_as_hash['single_image_masking_percentile']
        xpcs_settings['single_image_masking_log'] = xpcs_settings_as_hash['single_image_masking_log']
        xpcs_settings['numBins'] = xpcs_settings_as_hash['numBins']
        xpcs_settings['qBin_value'] = xpcs_settings_as_hash['qBin_value']
        xpcs_settings['qBin_range'] = xpcs_settings_as_hash['qBin_range']
        xpcs_settings['photonEnergy'] = xpcs_settings_as_hash['photonEnergy']
        
        if xpcs_settings_as_hash['litFramesList'] == -1:
            print("no data from lit frame finder, defaulting to manual array")
            #xpcs_settings["litFramesList"] = np.array([65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87])
            xpcs_settings["litFramesList"] = np.arange(23)+65
        else:
            print("using list frame list from lit frame finder")
            xpcs_settings["litFramesList"] = np.asarray(xpcs_settings_as_hash['litFramesList'])
            
        print(xpcs_settings["litFramesList"])
        # add or possibly overwrite XPCS settings with the received
        if self.xpcs_settings is None:
            self.xpcs_settings = xpcs_settings
        else:
            self.xpcs_settings.update(xpcs_settings)
            
        # make output
        strUpdateSettings = 'Settings update received:\n'    
        if 'photonEnergy' in xpcs_settings:
            photonEnergy = xpcs_settings['photonEnergy']
            strUpdateSettings += f'-- photonEnergy (keV)    : {photonEnergy}\n' 
            self.set('photonEnergy', photonEnergy)
        
        if 'qbin' in xpcs_settings:
            strUpdateSettings += f'-- q bin shape           : {xpcs_settings['qbin'].shape}\n' 

        if 'numBins' in xpcs_settings:
            numBins = xpcs_settings['numBins']
            strUpdateSettings += f'-- number of q bins      : {numBins}\n' 
            self.set('numBins', numBins)

        if 'single_image_masking' in xpcs_settings:
            self.set('single_image_masking', xpcs_settings['single_image_masking'])
            strUpdateSettings += f'-- single_image_masking  : {xpcs_settings["single_image_masking"]}\n' 

        if 'single_image_masking_percentile' in xpcs_settings:
            self.set('single_image_masking_percentile', xpcs_settings['single_image_masking_percentile'])
            strUpdateSettings += f'-- single_image_masking_percentile : {xpcs_settings["single_image_masking_percentile"]}\n' 

        if 'single_image_masking_log' in xpcs_settings:
            self.set('single_image_masking_log', xpcs_settings['single_image_masking_log'])
            strUpdateSettings += f'-- single_image_masking_log : {xpcs_settings["single_image_masking_log"]}\n' 

        if 'output_path' in xpcs_settings:
            myname = self.get('deviceId')
            onlinexpcsutils.write_data_to_file(self.online_processing_memory, xpcs_settings['output_path'], myname)
            self.log.WARN(f'writing data to file {xpcs_settings["output_path"]}')

        if 'clear_accumulating_memory' in xpcs_settings:
            self.online_processing_memory = onlinexpcsutils.clear_accumulating_processing_memory(self.online_processing_memory)
            self.log.WARN(f'Clearing Accumulating Memory')

        print(strUpdateSettings)
        '''
           
    def get_detector_data_list_from_matched_sources_and_the_gain_stage_data(self, sources):
        '''
        assemble matched sources into a list of pairs (module name, module data)
        of types:
        module name : string  e.g.  "Q1M3" for agipd
        module data : np array of shape (numBunches, SlowScanDim, FastScanDim)
        '''
        detector_data_list = list()
        
        # variable that will contain the sum of the np arrays that have the pixel count per gain stage
        numPixelPerGainStage = None
        for source, (data, timestamp) in sources.items():   
            
            unique_string_id = agipd_module_name_from_mid_agipd_source(source)

            self._shmem_handler.dereference_shmem_handles(data) 
            
            detectorData = data['image']['data']
            
            # ensure 3 dimensional
            if len(detectorData.shape) == 2:
                (slowDim, fastDim) = detectorData.shape
                detectorData.reshape(shape=(1,slowDim,fastDim))
            
            if numPixelPerGainStage is None:
                numPixelPerGainStage = data['numPixelsPerGainStage']
            else:
                numPixelPerGainStage += data['numPixelsPerGainStage']
            
            detector_data_list.append((unique_string_id, detectorData))
 
        return detector_data_list, numPixelPerGainStage
                    
    
    def preDestruction(self):
        print("CLOSING!")
        self.online_processing_memory = onlinexpcsutils.close_processing_plot_thread(self.online_processing_memory)
    
        self._reset()
        super().preDestruction()

    def print_data_processing_update(self, numBins, numBunches, trainId):
        
        date_time = datetime.today().strftime('%d/%m/%Y %H:%M:%S')
        

        my_device_id = self.getInstanceId()
        now = time.time()
        if now-self.timeSinceLastDataPrintUpdate > 300:
            print(f'{date_time} : {my_device_id} received data for trainid {trainId} with {numBins} qBins and {numBunches} bunches')
            self.timeSinceLastDataPrintUpdate = now

    my_generated_train_id = 21249888900
    def triggerSampleOutput(self):

        self.log.INFO("generating sample output")
        print("generating sample output")
        myGroupName = self.get('deviceId')[-2:]
        
        def generate_data(q):
            seconds = time.time()
            
            detectorData = onlinexpcsutils.get_test_data((50, 512, 128), q=q, photonEnergy=self.get('photonEnergy'))
            return detectorData
        
        
        for count in range(200):
            my_detector_data_list = list()        

            q = 0.1 + 0.08*math.sin(0.1*count)
            if myGroupName == "G1":
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(0) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(1) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(2) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(3) , generate_data(q)))
            if myGroupName == "G2":
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(4) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(5) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(6) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(7) , generate_data(q)))
            if myGroupName == "G3":
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(8)  , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(9)  , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(10) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(11) , generate_data(q)))
            if myGroupName == "G4":
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(12) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(13) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(14) , generate_data(q)))
                my_detector_data_list.append((onlinexpcsutils.agipd_source_module_no_to_unique_string_id(15) , generate_data(q)))
            
            OnlineMIDProcessor.my_generated_train_id += 1
            self.handle_detector_data_list(OnlineMIDProcessor.my_generated_train_id, my_detector_data_list)
            time.sleep(0.4) 

    
    def handle_detector_data_list(self, train_id, detector_data_list, numPixelPerGainStage=None):
         
       # time_to_handle_list_0 = time.perf_counter()
        
       # self.online_processing_memory, online_output_dict = onlinexpcsutils.compute_xpcs_from_detector_data_list(
       #     detector_data_list, 
       #     xpcs_processing_memory=self.online_processing_memory,
       #     xpcs_settings=self.xpcs_settings)       
       # print("time_to_handle_list 0", time.perf_counter() - time_to_handle_list_0)

        roi = self.xpcs_settings['qbin']
        self.runningstats_processing_memory, runningstats_output_dict = runningstats.process_detector_data_list(
            detector_data_list,
            roi,
            processing_memory = self.runningstats_processing_memory,
            lit_frames = self.xpcs_settings["litFramesList"]
            )
        
        if numPixelPerGainStage is not None:
            runningstats_output_dict['numPixelPerGainStage'] = numPixelPerGainStage


       # time_to_handle_list_1 = time.perf_counter()
        

       # selected_module_key = onlinexpcsutils.agipd_source_module_no_to_unique_string_id(self.get('module_number_for_plotting'))
        
       # print("time_to_handle_list 1", time.perf_counter() - time_to_handle_list_1)

        # get module related plots from processing memory
        # this is made via matplotlib, plotting takes time,
        # the following code ensure plotting is only asked every 0.9 seeconds
        
       # time_to_handle_list_2 = time.perf_counter()
        
       # self.online_processing_memory, current_image = onlinexpcsutils.xpcs_view_into_processing(
       #     self.online_processing_memory, 
       #     selected_module_key,
       #     train_id)
       # print("time_to_handle_list 2", time.perf_counter() - time_to_handle_list_2)


        # plot accumulating sum and lit pixel counter for panel data
       # time_to_handle_list_3 = time.perf_counter()
        
        #plot_hash = Hash()
        
#        npArray_sum = onlinexpcsutils.get_accumulating_sum(self.xpcs_processing_memory, selected_module_key)
#        if npArray_sum is not None:
#            plot_hash['data.accumulating_sum'] = ImageData(npArray_sum)
#        
#        npArray_lit_pixel_counter = onlinexpcsutils.get_accumulating_lit_pixel_counter(self.xpcs_processing_memory, selected_module_key)
#        if npArray_lit_pixel_counter is not None:
#            plot_hash['data.accumulating_lit_bunch'] = ImageData(npArray_lit_pixel_counter)

        #npArray_currentImage = onlinexpcsutils.get_current_Image(self.online_processing_memory, selected_module_key)
        #if npArray_currentImage is not None:
        #    plot_hash['data.current_det_image'] = ImageData(npArray_currentImage)



 #       if current_image is not None:
 #           plot_hash['data.masking_plot'] = ImageData(current_image)
        
        
       # print("time_to_handle_list 3", time.perf_counter() - time_to_handle_list_3)
        # send hash

       # time_to_handle_list_4 = time.perf_counter()

#        my_timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
#        myname = self.get('deviceId')
#        if self.assembled_output is not None:
#            self.assembled_output.write(
#                    plot_hash, ChannelMetaData(f'{myname}:plots', my_timestamp), copyAllData=False
#                )
#            self.assembled_output.update(safeNDArray=True)


       # print("time_to_handle_list _ 4", time.perf_counter() - time_to_handle_list_4)        
        
 #       time_to_handle_list_5 = time.perf_counter()

        
        output_hash = Hash() 
        if runningstats_output_dict is not None:
            output_hash = self.make_dict_to_hash(runningstats_output_dict)
        #xpcs_output_hash = Hash("output_dict", xpcs_output_dict)
        ## XPCS ANALYSIS END

        #######
        # karabo output
        my_timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        myname = self.get('deviceId')
        if self.output is not None:
            self.output.write(
                    output_hash, ChannelMetaData(f'{myname}:output', my_timestamp), copyAllData=False
                )
            self.output.update(safeNDArray=True)

       # print("time_to_handle_list _5", time.perf_counter() - time_to_handle_list_5)        
        
        if runningstats_output_dict is not None:
            if 'numBins' in runningstats_output_dict and 'numBunches' in runningstats_output_dict:
                self.print_data_processing_update(runningstats_output_dict['numBins'], runningstats_output_dict['numBunches'], train_id)
            else:
                self.print_data_processing_update(-1, -1, train_id)    
        else:
            self.print_data_processing_update(-1, -1, train_id)


    def on_matched_data(self, train_id, sources):

        # device processing timer start
        ts_start = default_timer()

        ########
        ## XPCS ANALYSIS START
        detector_data_list, numPixelPerGainStage = self.get_detector_data_list_from_matched_sources_and_the_gain_stage_data(sources)


        self.handle_detector_data_list(train_id, detector_data_list, numPixelPerGainStage=numPixelPerGainStage)
                
        # karabo bridge output
        #for source, (data, timestamp) in sources.items():
            #print("bridge output on:", source)
        #    self.zmq_output.write(source, data, timestamp)
            #self.zmq_output.write(source, xpcs_output, timestamp)
            #break
        #self.zmq_output.update()

        self._processing_time_tracker.update(default_timer() - ts_start)
        self.info["processingTime"] = self._processing_time_tracker.get() * 1000
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

        
